/*******************************************************************************
 * Copyright (c) 2010 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/
package org.openxma.dsl.core.resource;

import static com.google.common.collect.Iterables.transform;

import java.io.IOException;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.InternalEObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtext.naming.IQualifiedNameProvider;
import org.eclipse.xtext.resource.IEObjectDescription;
import org.eclipse.xtext.resource.IReferenceDescription;
import org.eclipse.xtext.resource.impl.DefaultReferenceDescription;
import org.eclipse.xtext.resource.impl.DefaultResourceDescription;
import org.openxma.dsl.core.model.ParameterDefinition;
import org.openxma.dsl.core.model.StatusFlag;
import org.openxma.dsl.core.util.CoreSwitch;

import com.google.common.base.Function;
import com.google.common.base.Predicate;
import com.google.common.base.Predicates;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;

/**
 * Backport index optimizations from Xtext 1.0.2/2.0.0 to Xtext currently used
 * version 1.0.2
 */
public class CoreDslResourceDescription extends DefaultResourceDescription {
    private static final Logger log = Logger.getLogger(CoreDslResourceDescription.class);
    private static final Predicate<EObject> EXPORTED_OBJECTS_FILTER = new ExportedObjectsFilter();
    private List<IReferenceDescription> referenceDescriptions;

    public CoreDslResourceDescription(Resource resource, IQualifiedNameProvider nameProvider) {
        super(resource, nameProvider);
    }

    @Override
    protected List<IEObjectDescription> computeExportedObjects() {
        if (!getResource().isLoaded()) {
            try {
                getResource().load(null);
            } catch (IOException e) {
                log.error(e.getMessage(), e);
                return Collections.<IEObjectDescription> emptyList();
            }
        }
        Iterable<EObject> contents = new Iterable<EObject>() {
            public Iterator<EObject> iterator() {
                return EcoreUtil.getAllProperContents(getResource(), true);
            }
        };
        Iterable<IEObjectDescription> result = transform(Iterables.filter(contents, getExportedObjectsFilter()),
                new Function<EObject, IEObjectDescription>() {
                    public IEObjectDescription apply(EObject from) {
                        return createIEObjectDescription(from);
                    }
                });
        Iterable<IEObjectDescription> filter = Iterables.filter(result, Predicates.notNull());
        return Lists.newArrayList(filter);
    }

    protected Predicate<EObject> getExportedObjectsFilter() {
        return EXPORTED_OBJECTS_FILTER;
    }

    @SuppressWarnings("unchecked")
    public Iterable<IReferenceDescription> getReferenceDescriptions() {
        if (this.referenceDescriptions == null) {
            Map<EObject, IEObjectDescription> eObject2exportedEObjects = createEObject2ExportedEObjectsMap(getExportedObjects());
            List<IReferenceDescription> referenceDescriptions = Lists.newArrayList();
            TreeIterator<EObject> contents = EcoreUtil.getAllProperContents(getResource(), true);
            while (contents.hasNext()) {
                EObject eObject = contents.next();
                EList<EReference> references = eObject.eClass().getEAllReferences();
                for (EReference eReference : references) {
                    if (!eReference.isContainment() && !eReference.isDerived() && !eReference.isVolatile()
                            && !eReference.isTransient() && eObject.eIsSet(eReference)) {
                        Object val = eObject.eGet(eReference);
                        if (val != null) {
                            if (eReference.isMany()) {
                                List<EObject> list = (List<EObject>) val;
                                for (int i = 0; i < list.size(); i++) {
                                    EObject to = list.get(i);
                                    if (isResolvedAndExternal(eObject, to)) {
                                        referenceDescriptions.add(new DefaultReferenceDescription(eObject, to,
                                                eReference, i, findExportedContainerURI(eObject,
                                                        eObject2exportedEObjects)));
                                    }
                                }
                            } else {
                                EObject to = (EObject) val;
                                if (isResolvedAndExternal(eObject, to)) {
                                    referenceDescriptions.add(new DefaultReferenceDescription(eObject, to, eReference,
                                            -1, findExportedContainerURI(eObject, eObject2exportedEObjects)));
                                }
                            }
                        }
                    }
                }
            }
            this.referenceDescriptions = referenceDescriptions;
        }
        return referenceDescriptions;
    }

    protected boolean isResolvedAndExternal(EObject from, EObject to) {
        return to != null && !((InternalEObject) to).eIsProxy() && from.eResource() != to.eResource();
    }

    private static class ExportedObjectsFilter extends CoreSwitch<Boolean> implements Predicate<EObject> {

        public boolean apply(EObject input) {
            return doSwitch(input);
        }

        @Override
        public Boolean defaultCase(EObject object) {
            return true;
        }

        @Override
        public Boolean caseParameterDefinition(ParameterDefinition object) {
            return false;
        }

        @Override
        public Boolean caseStatusFlag(StatusFlag object) {
            return false;
        }

    }
}
